"""Configuration for the run control."""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field

from gwsim_pop.config.logging import LoggingConfiguration
from gwsim_pop.config.output import OutputConfiguration


class RunConfiguration(BaseModel):
    """Configuration for the run control."""

    name: str = Field(default="simulation", description="Run identifier.")
    """Run identifier."""

    seed: int = Field(default=42, description="Global RNG seed (overridable via CLI).")
    """Global RNG seed (overridable via CLI)."""

    mode: Literal["fixed_n_samples", "duration"] = Field(
        default="duration", description="Mode of simulation. Supported: 'fixed_n_samples', 'duration'."
    )
    """Mode of simulation. Supported: 'fixed_n_samples', 'duration'."""

    n_samples: int = Field(
        default=1_000_000,
        description="Number of samples. Only used when 'mode' is 'fixed_n_samples'.",
    )
    """Number of samples. Only used when 'mode' is 'fixed_n_samples'."""

    duration: float = Field(
        default=1.0, description="Duration in year (365 days). This is used when 'mode' is 'duration'."
    )
    """Duration in year (365 days). This is used when 'mode' is 'duration'."""

    output: OutputConfiguration = Field(
        default_factory=OutputConfiguration, description="Configuration for the output."
    )
    """Configuration for the output."""

    logging: LoggingConfiguration = Field(
        default_factory=LoggingConfiguration, description="Configuration for the logging."
    )
    """Configuration for the logging."""



"""Configuration for the output."""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field


class OutputConfiguration(BaseModel):
    """Configuration for the output."""

    directory: str = Field(default=".", description="Output directory.")
    """Output directory."""

    format: Literal["hdf5", "npz", "csv"] = Field(default="hdf5", description="Storage format.")
    """Storage format."""

    compression: None | Literal["gzip", "lzf", "szip"] = Field(default="gzip", description="Compression filter.")
    """Compression filter."""

    overwrite: bool = Field(default=False, description="Whether or not to overwrite existing files.")
    """Whether or not to overwrite existing files."""

    save_metadata: bool = Field(default=True, description="Whether or not to save metadata.")
    """Whether or not to save metadata."""



"""Functions to smooth distributions."""

from __future__ import annotations

import jax
import jax.numpy as jnp
from jax import Array


def planck_tapering_function(x: Array, x_min: float, delta: float) -> Array:
    """Planck tapering function to smoothly transition between 0 and 1.

    References:
    - GWTC-4.0: Population Properties of Merging Compact Binaries: https://arxiv.org/abs/2508.18083

    Args:
        x: The input value.
        x_min: The minimum value of the transition region.
        delta: The width of the transition region.

    Returns:
        An array with the value of the tapering function at x.

    """

    def planck_rise(x: Array, x_min: float, delta: float) -> Array:
        """Planck rise function from 0 to 1 over the interval [x_min, x_min + delta].

        Args:
            x: The input value.
            x_min: The minimum value of the transition region.
            delta: The width of the transition region.

        Returns:
            The value of the rise function at x.

        """
        a = jnp.exp(delta / (x - x_min) + delta / (x - (x_min + delta)))
        return 1.0 / (1.0 + a)

    return jnp.where(
        x < x_min,
        0.0,
        jnp.where(
            x < x_min + delta,
            planck_rise(x, x_min, delta),
            1.0,
        ),
    )


def log_planck_tapering_function(x: Array, x_min: float, delta: float) -> Array:
    """Log of the Planck tapering function to smoothly transition between 0 and 1.

    References:
    - GWTC-4.0: Population Properties of Merging Compact Binaries: https://arxiv.org/abs/2508.18083

    Args:
        x: The input value.
        x_min: The minimum value of the transition region.
        delta: The width of the transition region.

    Returns:
        An array with the value of the log tapering function at x.

    """

    def log_planck_rise(x: Array, x_min: float, delta: float) -> Array:
        """Log of the Planck rise function from -inf to 0 over the interval [x_min, x_min + delta].

        Args:
            x: The input value.
            x_min: The minimum value of the transition region.
            delta: The width of the transition region.

        Returns:
            The log of the rise function value at x, in the range (-inf, 0].

        """
        exponent = delta / (x - x_min) + delta / (x - (x_min + delta))
        return -jax.nn.softplus(exponent)

    return jnp.where(
        x < x_min,
        -jnp.inf,
        jnp.where(
            x < x_min + delta,
            log_planck_rise(x, x_min, delta),
            0.0,
        ),
    )



"""Broken power law distribution implementation."""

from __future__ import annotations

import jax
import jax.numpy as jnp


def broken_power_law_unnormalized_logpdf(  # noqa: PLR0913
    x: jax.Array,
    x_min: float,
    x_max: float,
    transition: float,
    alpha_1: float,
    alpha_2: float,
) -> jax.Array:
    """Unnormalized log probability density function of a broken power law distribution.

    References:
    - GWTC-4.0: Population Properties of Merging Compact Binaries: https://arxiv.org/abs/2508.18083

    Args:
        x: The input value.
        x_min: The minimum value of the distribution.
        x_max: The maximum value of the distribution.
        transition: The transition point where the power law changes.
        alpha_1: The power law index for x < transition.
        alpha_2: The power law index for x >= transition.

    Returns:
        An array with the unnormalized log probability density at x.

    """
    return jnp.where(
        x < x_min,
        -jnp.inf,
        jnp.where(
            x < transition,
            -alpha_1 * jnp.log(x / transition),
            jnp.where(
                x < x_max,
                -alpha_2 * jnp.log(x / transition),
                -jnp.inf,
            ),
        ),
    )


def broken_power_law_log_normalization_constant(
    x_min: float,
    x_max: float,
    transition: float,
    alpha_1: float,
    alpha_2: float,
) -> jax.Array:
    """Log normalization constant of a broken power law distribution.

    References:
    - GWTC-4.0: Population Properties of Merging Compact Binaries: https://arxiv.org/abs/2508.18083

    Args:
        x_min: The minimum value of the distribution.
        x_max: The maximum value of the distribution.
        transition: The transition point where the power law changes.
        alpha_1: The power law index for x < transition.
        alpha_2: The power law index for x >= transition.

    Returns:
        The log normalization constant of the distribution.

    """
    return jnp.log(transition) + jnp.log(
        (1 - (x_min / transition) ** (1 - alpha_1)) / (1 - alpha_1)
        + ((x_max / transition) ** (1 - alpha_2) - 1) / (1 - alpha_2)
    )


def broken_power_law_logpdf(  # noqa: PLR0913
    x: jax.Array, x_min: float, x_max: float, transition: float, alpha_1: float, alpha_2: float
) -> jax.Array:
    """Log probability density function of a broken power law distribution.

    References:
    - GWTC-4.0: Population Properties of Merging Compact Binaries: https://arxiv.org/abs/2508.18083

    Args:
        x: The input value.
        x_min: The minimum value of the distribution.
        x_max: The maximum value of the distribution.
        transition: The transition point where the power law changes.
        alpha_1: The power law index for x < transition.
        alpha_2: The power law index for x >= transition.

    Returns:
        An array with the log probability density at x.

    """
    return broken_power_law_unnormalized_logpdf(
        x=x, x_min=x_min, x_max=x_max, transition=transition, alpha_1=alpha_1, alpha_2=alpha_2
    ) - broken_power_law_log_normalization_constant(
        x_min=x_min, x_max=x_max, transition=transition, alpha_1=alpha_1, alpha_2=alpha_2
    )



"""Configuration and fixtures for pytest."""

from __future__ import annotations

import pytest


@pytest.fixture
def some_name() -> str:
    """Provide a string name.

    Returns:
        A string name.

    """
    return "developer"



"""Tests for the cosmology configuration."""

from __future__ import annotations

from gwsim_pop.config.cosmology import CosmologicalArguments, CosmologyConfiguration


class TestCosmologicalArguments:
    """Test cases for CosmologicalArguments."""

    def test_default_values(self) -> None:
        """Test that default values are correctly set."""
        args = CosmologicalArguments()
        assert args.model_config == {"extra": "allow", "arbitrary_types_allowed": True, "validate_default": True}

    def test_extra_fields_allowed(self) -> None:
        """Test that extra fields are allowed."""
        args = CosmologicalArguments(extra_field="test_value")
        assert args.model_extra is not None
        assert args.model_extra.get("extra_field") == "test_value"

    def test_arbitrary_types_allowed(self) -> None:
        """Test that arbitrary types are allowed."""
        # The CosmologicalArguments is a generic class, so we test by passing various valid arguments
        args = CosmologicalArguments()
        assert args is not None


class TestCosmologyConfiguration:
    """Test cases for CosmologyConfiguration."""

    def test_default_values(self) -> None:
        """Test that default values are correctly set."""
        config = CosmologyConfiguration()
        assert config.model == "Planck18"
        assert isinstance(config.arguments, CosmologicalArguments)
        assert config.arguments.model_config == {
            "extra": "allow",
            "arbitrary_types_allowed": True,
            "validate_default": True,
        }

    def test_explicit_values(self) -> None:
        """Test that explicit values are correctly set."""
        config = CosmologyConfiguration(model="WMAP9", arguments=CosmologicalArguments())
        assert config.model == "WMAP9"
        assert isinstance(config.arguments, CosmologicalArguments)

    def test_arguments_default_factory(self) -> None:
        """Test that arguments uses the default factory correctly."""
        config = CosmologyConfiguration()
        assert isinstance(config.arguments, CosmologicalArguments)
        # Should have the default model_config
        assert config.arguments.model_config == {
            "extra": "allow",
            "arbitrary_types_allowed": True,
            "validate_default": True,
        }

    def test_model_field_description(self) -> None:
        """Test that model field has correct description."""
        config = CosmologyConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["model"].description == "Cosmological model."

    def test_arguments_field_description(self) -> None:
        """Test that arguments field has correct description."""
        config = CosmologyConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["arguments"].description == "Arguments for the model."

    def test_configuration_from_dict(self) -> None:
        """Test that configuration can be created from a dictionary."""
        config_dict = {"model": "WMAP9", "arguments": {}}
        config = CosmologyConfiguration(**config_dict)
        assert config.model == "WMAP9"
        assert isinstance(config.arguments, CosmologicalArguments)

    def test_nested_configuration(self) -> None:
        """Test nested configuration with arguments."""
        args = CosmologicalArguments()
        config = CosmologyConfiguration(model="CustomModel", arguments=args)
        assert config.model == "CustomModel"
        assert isinstance(config.arguments, CosmologicalArguments)

    def test_empty_arguments(self) -> None:
        """Test configuration with empty arguments."""
        config = CosmologyConfiguration(arguments=CosmologicalArguments())
        assert config.model == "Planck18"
        assert isinstance(config.arguments, CosmologicalArguments)
        # Check that arguments can accept extra fields
        extra_args = CosmologicalArguments(extra_test="value")
        assert extra_args.model_extra is not None
        assert extra_args.model_extra.get("extra_test") == "value"



"""Utility functions for the gwsim_pop package."""

from __future__ import annotations

from gwsim_pop.utils.log import get_version_information, setup_logger

__all__ = ["get_version_information", "setup_logger"]



"""Configuration for logging."""

from __future__ import annotations

from pydantic import BaseModel, Field

from gwsim_pop.utils.log import LoggingLevel


class LoggingConfiguration(BaseModel):
    """Configuration for the logging."""

    level: LoggingLevel = Field(default=LoggingLevel.INFO, description="Logging level.")
    """Logging level."""

    filename: str = Field(default="simulation.log", description="Name of the log file.")
    """Name of the log file."""



"""Tests for logging utility functions."""

from __future__ import annotations

import logging
import tempfile
from contextlib import contextmanager
from pathlib import Path

import pytest

from gwsim_pop.utils.log import LoggingLevel, get_version_information, setup_logger


@contextmanager
def _temp_file_logger(log_file: Path, level=logging.INFO):
    """Context manager: adds & removes FileHandler cleanly."""
    logger = logging.getLogger("test_logger")  # ← use a dedicated name to avoid pollution
    handler = logging.FileHandler(log_file, encoding="utf-8")
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))  # your format

    logger.addHandler(handler)
    logger.setLevel(level)

    try:
        yield handler, log_file  # or just yield log_file if you don't need the handler
    finally:
        # Critical cleanup — do this BEFORE temp dir is removed
        handler.flush()
        handler.close()
        logger.removeHandler(handler)
        # Optional: logger.setLevel(logging.NOTSET) or remove if you reset everything


@pytest.fixture
def temp_log_file(tmp_path: Path):
    """Fixture: provides a temp directory + sets up logging to test.log inside it.

    Automatically cleans up the FileHandler so Windows doesn't complain.
    """
    log_path = tmp_path / "test.log"

    with _temp_file_logger(log_path) as (_handler, path):
        yield path  # ← tests receive only the Path to the log file


@pytest.fixture(autouse=True)
def cleanup_logger() -> None:
    """Clean up logger state before and after each test."""
    logger = logging.getLogger("gwsim_pop")
    logger.handlers.clear()
    logger.setLevel(logging.NOTSET)


class TestLoggingLevel:
    """Test LoggingLevel enum."""

    def test_enum_values(self) -> None:
        """Test that enum values match logging module."""
        assert LoggingLevel.NOTSET.value == "NOTSET"
        assert LoggingLevel.DEBUG.value == "DEBUG"
        assert LoggingLevel.INFO.value == "INFO"
        assert LoggingLevel.WARNING.value == "WARNING"
        assert LoggingLevel.ERROR.value == "ERROR"
        assert LoggingLevel.CRITICAL.value == "CRITICAL"

    def test_enum_is_str(self) -> None:
        """Test that LoggingLevel is a string enum."""
        assert isinstance(LoggingLevel.INFO, str)
        assert LoggingLevel.INFO == "INFO"


class TestGetVersionInformation:
    """Test get_version_information function."""

    def test_returns_string(self) -> None:
        """Test that function returns a string."""
        version = get_version_information()
        assert isinstance(version, str)
        assert len(version) > 0


class TestSetupLogger:
    """Test setup_logger function."""

    def test_default_parameters(self) -> None:
        """Test setup_logger with default parameters."""
        setup_logger()
        logger = logging.getLogger("gwsim_pop")
        assert logger.level == logging.INFO

    def test_custom_log_level_enum(self) -> None:
        """Test setup_logger with LoggingLevel enum."""
        setup_logger(log_level=LoggingLevel.DEBUG)
        logger = logging.getLogger("gwsim_pop")
        assert logger.level == logging.DEBUG

    def test_custom_log_level_string(self) -> None:
        """Test setup_logger with log level as string."""
        setup_logger(log_level="WARNING")
        logger = logging.getLogger("gwsim_pop")
        assert logger.level == logging.WARNING

    def test_custom_log_level_integer(self) -> None:
        """Test setup_logger with log level as integer."""
        setup_logger(log_level=logging.ERROR)
        logger = logging.getLogger("gwsim_pop")
        assert logger.level == logging.ERROR

    def test_invalid_log_level_string(self) -> None:
        """Test setup_logger with invalid log level string."""
        with pytest.raises(ValueError, match="log_level invalid_level not understood"):
            setup_logger(log_level="invalid_level")

    def test_no_file_handler_without_filename(self) -> None:
        """Test that no file handler is created without filename."""
        setup_logger(filename=None)
        logger = logging.getLogger("gwsim_pop")
        file_handlers = [h for h in logger.handlers if isinstance(h, logging.FileHandler)]
        assert len(file_handlers) == 0

    def test_file_handler_with_filename(self, temp_log_file) -> None:
        """Test that file handler is created with filename."""
        setup_logger(outdir=temp_log_file.parent, filename=temp_log_file.name)
        logger = logging.getLogger("gwsim_pop")
        file_handlers = [h for h in logger.handlers if isinstance(h, logging.FileHandler)]
        assert len(file_handlers) == 1
        assert temp_log_file.exists()

    def test_creates_output_directory(self) -> None:
        """Test that output directory is created if it doesn't exist."""
        with tempfile.TemporaryDirectory() as tmpdir:
            new_dir = Path(tmpdir) / "new_subdir"
            setup_logger(outdir=str(new_dir), filename="test.log")
            logger = logging.getLogger("gwsim_pop")
            file_handlers = [h for h in logger.handlers if isinstance(h, logging.FileHandler)]
            assert len(file_handlers) == 1
            assert new_dir.exists()

            if file_handlers:
                handler = file_handlers[0]
                handler.flush()  # ensure all writes are done
                handler.close()  # closes the file descriptor
                logger.removeHandler(handler)  # detach from logger

    def test_stream_handler_added(self) -> None:
        """Test that stream handler is added."""
        setup_logger()
        logger = logging.getLogger("gwsim_pop")
        stream_handlers = [
            h
            for h in logger.handlers
            if isinstance(h, logging.StreamHandler) and not isinstance(h, logging.FileHandler)
        ]
        assert len(stream_handlers) >= 1

    def test_no_duplicate_stream_handler(self) -> None:
        """Test that duplicate stream handlers are not added."""
        setup_logger()
        setup_logger()
        logger = logging.getLogger("gwsim_pop")
        stream_handlers = [
            h
            for h in logger.handlers
            if isinstance(h, logging.StreamHandler) and not isinstance(h, logging.FileHandler)
        ]
        assert len(stream_handlers) == 1

    def test_no_duplicate_file_handler(self, temp_log_file) -> None:
        """Test that duplicate file handlers are not added."""
        setup_logger(outdir=temp_log_file.parent, filename=temp_log_file.name)
        setup_logger(outdir=temp_log_file.parent, filename=temp_log_file.name)
        logger = logging.getLogger("gwsim_pop")
        file_handlers = [h for h in logger.handlers if isinstance(h, logging.FileHandler)]
        assert len(file_handlers) == 1

    def test_print_version(self, temp_log_file) -> None:
        """Test that version is printed when print_version=True."""
        setup_logger(outdir=temp_log_file.parent, filename=temp_log_file.name, print_version=True)
        _logger = logging.getLogger("gwsim_pop")
        content = temp_log_file.read_text()
        assert "Running gwsim_pop version:" in content

    def test_log_message_format(self, temp_log_file) -> None:
        """Test that log messages have correct format."""
        setup_logger(outdir=temp_log_file.parent, filename=temp_log_file.name, log_level=logging.INFO)
        logger = logging.getLogger("gwsim_pop")
        logger.info("Test message")
        content = temp_log_file.read_text()
        assert "Test message" in content
        assert "INFO" in content



"""Main entry point for the gwsim_pop package."""

from __future__ import annotations

if __name__ == "__main__":
    from gwsim_pop.utils.log import setup_logger

    setup_logger(print_version=True)



"""Tests for post-processing configuration."""

import pytest
from pydantic import ValidationError

from gwsim_pop.config.post_processing import HookArguments, HookConfiguration, PostProcessingConfiguration


class TestHookArguments:
    """Tests for HookArguments class."""

    def test_initialization(self):
        """Test that HookArguments can be initialized."""
        args = HookArguments()
        assert isinstance(args, HookArguments)


class TestHookConfiguration:
    """Tests for HookConfiguration class."""

    def test_initialization_without_arguments(self):
        """Test that HookConfiguration can be initialized without arguments."""
        config = HookConfiguration(name="test_hook")
        assert config.name == "test_hook"
        assert isinstance(config.arguments, HookArguments)

    def test_initialization_with_arguments(self):
        """Test that HookConfiguration can be initialized with arguments."""
        args = HookArguments()
        config = HookConfiguration(name="test_hook", arguments=args)
        assert config.name == "test_hook"
        assert isinstance(config.arguments, HookArguments)

    def test_name_validation(self):
        """Test that name validation works correctly."""
        with pytest.raises(ValidationError):
            HookConfiguration()


class TestPostProcessingConfiguration:
    """Tests for PostProcessingConfiguration class."""

    def test_initialization(self):
        """Test that PostProcessingConfiguration can be initialized."""
        config = PostProcessingConfiguration()
        assert isinstance(config, PostProcessingConfiguration)
        assert config.hooks == []

    def test_hooks_validation_empty(self):
        """Test that empty hooks list passes validation."""
        config = PostProcessingConfiguration(hooks=[])
        assert config.hooks == []

    def test_hooks_validation_non_empty_raises_error(self):
        """Test that non-empty hooks list raises validation error."""
        with pytest.raises(ValidationError, match=r"Post-processing hook is not implemented yet"):
            PostProcessingConfiguration(hooks=[HookConfiguration(name="test_hook")])



"""Utility functions for logging."""

from __future__ import annotations

import logging
from enum import Enum
from pathlib import Path

from gwsim_pop.version import __version__


class LoggingLevel(str, Enum):
    """Logging level."""

    NOTSET = "NOTSET"
    """When set on a logger, indicates that ancestor loggers are to be consulted to determine the effective level.
    If that still resolves to NOTSET, then all events are logged. When set on a handler, all events are handled."""

    DEBUG = "DEBUG"
    """Detailed information, typically only of interest to a developer trying to diagnose a problem."""

    INFO = "INFO"
    """Confirmation that things are working as expected."""

    WARNING = "WARNING"
    """An indication that something unexpected happened, or that a problem might occur in the near future (e.g. 'disk space low'). The software is still working as expected."""

    ERROR = "ERROR"
    """Due to a more serious problem, the software has not been able to perform some function."""

    CRITICAL = "CRITICAL"
    """A serious error, indicating that the program itself may be unable to continue running."""


def get_version_information() -> str:
    """Get the version information.

    Returns:
        str: Version information.

    """
    return __version__


def setup_logger(
    outdir: str = ".",
    filename: str | None = None,
    log_level: LoggingLevel | str | int = "INFO",
    print_version: bool = False,
) -> None:
    """Set up logging output: call at the start of the script to use.

    Args:
        outdir: Output directory for log file.
        filename: Name of the log file. If None, no log file is created.
        log_level: Logging level as string or integer.
        print_version: Whether to print version information to the log.

    """
    if isinstance(log_level, LoggingLevel):
        level = log_level.value
    elif isinstance(log_level, str):
        try:
            level = getattr(logging, log_level.upper())
        except AttributeError as e:
            raise ValueError(f"log_level {log_level} not understood") from e
    else:
        level = int(log_level)

    logger = logging.getLogger("gwsim_pop")
    logger.propagate = False
    logger.setLevel(level)

    if not any(
        isinstance(h, logging.StreamHandler) and not isinstance(h, logging.FileHandler) for h in logger.handlers
    ):
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(
            logging.Formatter("%(asctime)s %(name)s %(levelname)-8s: %(message)s", datefmt="%H:%M")
        )
        stream_handler.setLevel(level)
        logger.addHandler(stream_handler)

    if filename:
        outdir_path = Path(outdir)
        outdir_path.mkdir(parents=True, exist_ok=True)
        log_file = outdir_path / filename
        if not any(
            isinstance(h, logging.FileHandler) and Path(h.baseFilename).resolve() == log_file.resolve()
            for h in logger.handlers
        ):
            file_handler = logging.FileHandler(log_file, mode="a")
            file_handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)-8s: %(message)s", datefmt="%H:%M"))
            file_handler.setLevel(level)
            logger.addHandler(file_handler)

    for handler in logger.handlers:
        handler.setLevel(level)

    if print_version:
        version = get_version_information()
        logger.info("Running gwsim_pop version: %s", version)



"""Random number generator manager."""

from __future__ import annotations

import logging
import secrets
from pathlib import Path

import jax
import jax.numpy as jnp
from jax import Array

logger = logging.getLogger("gwsim_pop")


class RNGManager:
    """Manager for random number generation with state persistence."""

    def __init__(self, seed: int | None = None) -> None:
        """Initialize the RNG manager.

        Args:
            seed: Random seed. If None, draw a random integer between [0, 2**63).

        """
        self._seed = seed
        self._key = jax.random.key(secrets.randbelow(2**63) if seed is None else seed)
        self._saved_keys: list[Array] = []

    def __repr__(self) -> str:
        """Return the string representation.

        Returns:
            String representation.

        """
        return f"RNGManager(seed={self._seed})"

    def __enter__(self) -> RNGManager:
        """Enter the context manager.

        Use the context manager when keeping the initial state of the RNG is preferred.

        Returns:
            RNGManager.

        """
        self._saved_keys.append(self._key)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the context manager."""
        self._key = self._saved_keys.pop()

    @property
    def key(self) -> Array:
        """Get the random number generator key.

        Returns:
            JAX random key.

        """
        return self._key

    @key.setter
    def key(self, value: Array) -> None:
        """Set the random number generator key.

        Args:
            value: The new key value as a jax.Array. If the dtype starts with
                ``key<``, it is used as-is. If the dtype is ``uint32``, it is
                wrapped into a PRNG key via ``jax.random.wrap_key_data``.

        Raises:
            TypeError: If the value is not a jax.Array or has an unsupported dtype.

        """
        if not isinstance(value, Array):
            raise ValueError("value has to be jax.Array.")
        value_dtype_str = str(value.dtype)
        if value_dtype_str.startswith("key<"):
            self._key = value
        elif value_dtype_str == "uint32":
            self._key = jax.random.wrap_key_data(value)
        else:
            raise TypeError(f"Unexpected dtype of value={value_dtype_str}.")

    @property
    def key_data(self) -> Array:
        """Get the key data.

        The function can be used to retrieve the key data for storing in disk.

        Returns:
            Key data.

        """
        return jax.random.key_data(self._key)

    @key_data.setter
    def key_data(self, value: Array) -> None:
        """Set the key by key data.

        Args:
            value: Key data.

        """
        self._key = jax.random.wrap_key_data(value)

    @property
    def new_key(self) -> Array:
        """Get a new key.

        Returns:
            A new key.

        """
        self._key, sub_key = jax.random.split(key=self._key)
        return sub_key

    def save_key(self, path: str | Path) -> None:
        """Save RNG key to file.

        Args:
            path: Path to save the key file.

        """
        path = Path(path)
        if path.suffix != ".npy":
            logger.warning(f"The suffix of path={path} is not '.npy'.")
            path = path.with_suffix(".npy")
            logger.warning(f"The path is updated to {path}.")
        jnp.save(file=path, arr=self.key_data)

    def load_key(self, path: str | Path) -> None:
        """Load RNG key from file.

        Args:
            path: Path to load the key file.

        """
        path = Path(path)
        if path.suffix != ".npy":
            logger.warning(f"The suffix of path={path} is not '.npy'.")
            path = path.with_suffix(".npy")
            logger.warning(f"The path is updated to {path}.")
        key_data = jnp.load(path)
        self.key_data = key_data



"""Core functions for distributions module."""

from __future__ import annotations

from gwsim_pop.core.distributions.broken_power_law import (
    broken_power_law_log_normalization_constant,
    broken_power_law_logpdf,
    broken_power_law_unnormalized_logpdf,
)
from gwsim_pop.core.distributions.smoothing import planck_tapering_function

__all__ = [
    "broken_power_law_log_normalization_constant",
    "broken_power_law_logpdf",
    "broken_power_law_unnormalized_logpdf",
    "planck_tapering_function",
]



"""Configuration for the advanced features."""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field, field_validator


class AdvancedConfiguration(BaseModel):
    """Configuration for the advanced features."""

    backend: Literal["jax", "numpy"] = Field(default="jax", description="Backend for the computation.")
    """Backend for the computation."""

    jit: bool = Field(default=True, description="Whether to jit critical functions.")
    """Whether to jit critical functions."""

    vectorization: Literal["vmap", "manual loop", "pmap"] = Field(default="vmap", description="Vectorization method.")
    """Vectorization method."""

    device: Literal["cpu", "gpu", "tpu", "auto"] = Field(default="auto", description="Device to use for the run.")
    """Device to use for the run."""

    @field_validator("backend")
    @classmethod
    def validate_backend(cls, value: Literal["jax", "numpy"]) -> Literal["jax", "numpy"]:
        """Validate the backend to use.

        Args:
            value: Name of the backend.

        Returns:
            Name of the validated backend.

        Raises:
            ValueError: If value is numpy.
        """
        if value != "jax":
            raise ValueError(f"backend='{value}' is not implemented yet.")
        return value



"""Tests for selection configuration."""

import pytest
from pydantic import ValidationError

from gwsim_pop.config.selection import SelectionConfiguration


class TestSelectionConfiguration:
    """Tests for SelectionConfiguration class."""

    def test_initialization(self):
        """Test that SelectionConfiguration can be initialized."""
        config = SelectionConfiguration()
        assert isinstance(config, SelectionConfiguration)

    def test_default_values(self):
        """Test that default values are set correctly."""
        config = SelectionConfiguration()
        assert config.enabled is False
        assert config.method == "snr"
        assert config.arguments == {}

    def test_initialization_with_default_values(self):
        """Test that SelectionConfiguration can be initialized with default values."""
        config = SelectionConfiguration()
        assert config.enabled is False
        assert config.method == "snr"
        assert config.arguments == {}

    def test_enabled_validation_raises_error(self):
        """Test that enabling selection raises validation error."""
        with pytest.raises(ValidationError, match="The selection feature is not available yet"):
            SelectionConfiguration(enabled=True)

    def test_enabled_false_works(self):
        """Test that setting enabled to False works."""
        config = SelectionConfiguration(enabled=False)
        assert config.enabled is False

    def test_initialization_with_arguments(self):
        """Test that SelectionConfiguration can be initialized with arguments."""
        config = SelectionConfiguration(arguments={"threshold": 5.0})
        assert config.enabled is False
        assert config.arguments == {"threshold": 5.0}



"""Configuration for the population."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field


class PopulationArguments(BaseModel):
    """Population arguments."""

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True, validate_default=True)
    """Allow additional arguments."""


class PopulationConfiguration(BaseModel):
    """Configuration for the population."""

    model: str = Field(..., description="Name of the model.")
    """Name of the model."""

    arguments: PopulationArguments = Field(
        default_factory=PopulationArguments, description="Arguments for the population."
    )
    """Arguments for the population."""



"""Tests for main configuration."""

import tempfile
from pathlib import Path

import pytest
from pydantic import ValidationError

from gwsim_pop.config.main import MainConfiguration


class TestMainConfiguration:
    """Tests for MainConfiguration class."""

    def test_initialization(self):
        """Test that MainConfiguration can be initialized."""
        config = MainConfiguration()
        assert isinstance(config, MainConfiguration)
        assert config.config_version == "0.1.0"

    def test_default_values(self):
        """Test that default values are set correctly."""
        config = MainConfiguration()
        assert config.config_version == "0.1.0"
        assert hasattr(config, "run")
        assert hasattr(config, "cosmology")
        assert hasattr(config, "selection")
        assert hasattr(config, "post_processing")
        assert hasattr(config, "advanced")

    def test_initialization_with_values(self):
        """Test that MainConfiguration can be initialized with specific values."""
        config = MainConfiguration(config_version="1.0.0")
        assert config.config_version == "1.0.0"

    def test_from_file_method(self):
        """Test that from_file method works correctly."""
        # Create a temporary YAML file with minimal configuration
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write("""
config_version: "1.0.0"
run:
  name: test_run
""")
            temp_file = f.name

        try:
            config = MainConfiguration.from_file(temp_file)
            assert config.config_version == "1.0.0"
            assert config.run.name == "test_run"
        finally:
            Path(temp_file).unlink()

    def test_to_file_method(self):
        """Test that to_file method works correctly."""
        config = MainConfiguration(config_version="1.0.0")

        # Create a temporary file to write to
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            temp_file = f.name

        try:
            config.to_file(temp_file)
            # Verify file was created
            assert Path(temp_file).exists()
        finally:
            Path(temp_file).unlink()

    def test_extra_fields_forbidden(self):
        """Test that extra fields are forbidden in MainConfiguration."""
        # This should raise a ValidationError when trying to create a config with extra field
        with pytest.raises(ValidationError):
            MainConfiguration(extra_field="should_not_be_allowed")



"""Tests for RNGManager."""

from __future__ import annotations

from pathlib import Path

import jax
import jax.numpy as jnp
import pytest

from gwsim_pop.core.rng import RNGManager


class TestRNGManager:
    """Test suite for RNGManager."""

    def test_init_with_seed(self) -> None:
        """Test initialization with a fixed seed."""
        rng = RNGManager(seed=42)
        assert rng._seed == 42  # noqa: PLR2004
        assert isinstance(rng.key, jax.Array)

    def test_init_without_seed(self) -> None:
        """Test initialization without a seed."""
        rng = RNGManager(seed=None)
        assert rng._seed is None
        assert isinstance(rng.key, jax.Array)

    def test_key_property(self) -> None:
        """Test that key property returns a jax Array."""
        rng = RNGManager(seed=123)
        assert isinstance(rng.key, jax.Array)

    def test_key_property_setter(self) -> None:
        """Test setting RNG key."""
        rng1 = RNGManager(seed=456)
        key = rng1.key

        rng2 = RNGManager(seed=789)
        rng2.key = key

        assert jnp.array_equal(rng1.key, rng2.key)

    def test_save_key(self, tmp_path: Path) -> None:
        """Test saving RNG key to file."""
        rng = RNGManager(seed=111)

        output_path = tmp_path / "rng_key.npy"
        rng.save_key(output_path)

        assert Path(output_path).exists()

    def test_load_key(self, tmp_path: Path) -> None:
        """Test loading RNG key from file."""
        rng1 = RNGManager(seed=222)

        output_path = tmp_path / "rng_key_load.npy"
        rng1.save_key(output_path)

        rng2 = RNGManager(seed=333)
        rng2.load_key(output_path)

        assert jnp.array_equal(rng1.key, rng2.key)

    def test_reproducibility_with_seed(self) -> None:
        """Test that same seed produces same random numbers."""
        rng1 = RNGManager(seed=444)
        rng2 = RNGManager(seed=444)

        _, subkey1 = jax.random.split(rng1.key)
        _, subkey2 = jax.random.split(rng2.key)

        val1 = jax.random.uniform(subkey1)
        val2 = jax.random.uniform(subkey2)

        assert val1 == val2

    def test_different_seeds_produce_different_values(self) -> None:
        """Test that different seeds produce different random numbers."""
        rng1 = RNGManager(seed=555)
        rng2 = RNGManager(seed=666)

        _, subkey1 = jax.random.split(rng1.key)
        _, subkey2 = jax.random.split(rng2.key)

        val1 = jax.random.uniform(subkey1)
        val2 = jax.random.uniform(subkey2)

        assert val1 != val2

    def test_key_persistence_across_calls(self) -> None:
        """Test that RNG key persists between calls."""
        rng = RNGManager(seed=777)

        key1, subkey1 = jax.random.split(rng.key)
        _, subkey2 = jax.random.split(key1)

        val1 = jax.random.uniform(subkey1)
        val2 = jax.random.uniform(subkey2)

        assert val1 != val2

    def test_save_key_with_string_path(self, tmp_path: Path) -> None:
        """Test saving key with string path."""
        rng = RNGManager(seed=888)
        output_path = str(tmp_path / "rng_key_str.npy")

        rng.save_key(output_path)

        assert Path(output_path).exists()

    def test_load_key_with_string_path(self, tmp_path: Path) -> None:
        """Test loading key with string path."""
        rng1 = RNGManager(seed=999)
        output_path = str(tmp_path / "rng_key_load_str.npy")

        rng1.save_key(output_path)

        rng2 = RNGManager(seed=101)
        rng2.load_key(output_path)

        assert jnp.array_equal(rng1.key, rng2.key)

    def test_multiple_save_load_cycles(self, tmp_path: Path) -> None:
        """Test multiple save and load cycles."""
        rng = RNGManager(seed=121)

        for i in range(3):
            key_before = rng.key
            rng.save_key(tmp_path / f"key_cycle_{i}.npy")
            rng.load_key(tmp_path / f"key_cycle_{i}.npy")
            key_after = rng.key

            assert jnp.array_equal(key_before, key_after)

    def test_key_generates_different_values_on_each_call(self) -> None:
        """Test that RNG generates different values on successive calls."""
        rng = RNGManager(seed=131)

        key = rng.key
        values = []
        for _ in range(5):
            key, subkey = jax.random.split(key)
            values.append(jax.random.uniform(subkey))

        assert len(values) == 5  # noqa: PLR2004
        assert all(0 <= v < 1 for v in values)

    def test_key_can_produce_integers(self) -> None:
        """Test that RNG can generate integers."""
        rng = RNGManager(seed=141)

        _, subkey = jax.random.split(rng.key)
        val = jax.random.randint(subkey, shape=(), minval=0, maxval=100)

        assert isinstance(val, jax.Array)
        assert 0 <= val < 100  # noqa: PLR2004

    def test_key_can_produce_normal_distribution(self) -> None:
        """Test that RNG can generate normal distribution."""
        rng = RNGManager(seed=151)

        _, subkey = jax.random.split(rng.key)
        values = jax.random.normal(subkey, shape=(100,))

        assert len(values) == 100  # noqa: PLR2004
        assert isinstance(values, jax.Array)

    def test_repr_with_seed(self) -> None:
        """Test __repr__ with a seed."""
        rng = RNGManager(seed=42)

        repr_str = repr(rng)

        assert repr_str == "RNGManager(seed=42)"

    def test_repr_without_seed(self) -> None:
        """Test __repr__ without a seed."""
        rng = RNGManager(seed=None)

        repr_str = repr(rng)

        assert repr_str == "RNGManager(seed=None)"

    def test_context_manager_saves_and_restores_key(self) -> None:
        """Test that context manager saves and restores RNG key."""
        rng = RNGManager(seed=123)
        key_before = rng.key

        with rng:
            # Generate a new key inside context
            new_key1 = rng.new_key
            new_key2 = rng.new_key

            # Keys should be different from original
            assert not jnp.array_equal(key_before, new_key1)
            assert not jnp.array_equal(key_before, new_key2)
            assert not jnp.array_equal(new_key1, new_key2)

        # After context, key should be restored to original
        assert jnp.array_equal(key_before, rng.key)

    def test_context_manager_with_exception(self) -> None:
        """Test that context manager restores key even when exception occurs."""
        rng = RNGManager(seed=456)
        key_before = rng.key

        def func(rng):
            # Generate a new key
            _ = rng.new_key

            # Raise an exception
            raise ValueError("Test exception")

        with pytest.raises(ValueError, match="Test exception"), rng:
            func(rng)

        # After exception, key should still be restored
        assert jnp.array_equal(key_before, rng.key)

    def test_key_setter_with_invalid_type(self) -> None:
        """Test key setter raises ValueError with invalid type."""
        rng = RNGManager(seed=789)

        with pytest.raises(ValueError, match=r"value has to be jax.Array."):
            rng.key = "invalid"

    def test_key_data_property_getter(self) -> None:
        """Test key_data property getter."""
        rng = RNGManager(seed=101)
        key_data = rng.key_data

        assert isinstance(key_data, jax.Array)
        # Key data should be a 2-element array for PRNG keys
        assert key_data.shape == (2,)

    def test_key_data_property_setter(self) -> None:
        """Test key_data property setter."""
        rng1 = RNGManager(seed=202)
        key_data = rng1.key_data

        rng2 = RNGManager(seed=303)
        rng2.key_data = key_data

        assert jnp.array_equal(rng1.key, rng2.key)

    def test_new_key_property(self) -> None:
        """Test new_key property updates internal key and produces different keys."""
        rng = RNGManager(seed=404)

        key1 = rng.new_key
        key2 = rng.new_key
        key3 = rng.new_key

        # All keys should be different
        assert not jnp.array_equal(key1, key2)
        assert not jnp.array_equal(key2, key3)
        assert not jnp.array_equal(key1, key3)

    def test_save_key_with_invalid_path(self, tmp_path: Path) -> None:
        """Test saving key with invalid path raises error."""
        rng = RNGManager(seed=505)
        invalid_path = tmp_path / "nonexistent_dir" / "rng_key.npy"

        with pytest.raises(OSError, match="No such file or directory"):
            rng.save_key(invalid_path)

    def test_load_key_with_invalid_path(self) -> None:
        """Test loading key with invalid path raises error."""
        rng = RNGManager(seed=606)
        invalid_path = Path("/nonexistent/path/rng_key.npy")

        with pytest.raises(FileNotFoundError):
            rng.load_key(invalid_path)

    def test_load_key_with_corrupted_file(self, tmp_path: Path) -> None:
        """Test loading key with corrupted file raises error."""
        rng = RNGManager(seed=707)
        corrupted_path = tmp_path / "corrupted_key.npy"

        # Write invalid data
        with open(corrupted_path, "w") as f:
            f.write("not a valid numpy file")

        with pytest.raises(Exception, match=r"This file contains pickled"):
            rng.load_key(corrupted_path)



"""Tests for run configuration."""

from gwsim_pop.config.run import RunConfiguration


class TestRunConfiguration:
    """Tests for RunConfiguration class."""

    def test_initialization(self):
        """Test that RunConfiguration can be initialized."""
        config = RunConfiguration()
        assert isinstance(config, RunConfiguration)

    def test_default_values(self):
        """Test that default values are set correctly."""
        config = RunConfiguration()
        assert config.name == "simulation"
        assert config.seed == 42  # noqa: PLR2004
        assert config.mode == "duration"
        assert config.n_samples == 1_000_000  # noqa: PLR2004
        assert config.duration == 1.0

    def test_initialization_with_values(self):
        """Test that RunConfiguration can be initialized with specific values."""
        config = RunConfiguration(name="test_run", seed=123, mode="fixed_n_samples", n_samples=500000, duration=2.0)
        assert config.name == "test_run"
        assert config.seed == 123  # noqa: PLR2004
        assert config.mode == "fixed_n_samples"
        assert config.n_samples == 500000  # noqa: PLR2004
        assert config.duration == 2.0  # noqa: PLR2004

    def test_mode_validation(self):
        """Test that mode validation works correctly."""
        # Test valid modes
        config1 = RunConfiguration(mode="fixed_n_samples")
        config2 = RunConfiguration(mode="duration")
        assert config1.mode == "fixed_n_samples"
        assert config2.mode == "duration"

    def test_output_and_logging_defaults(self):
        """Test that output and logging configurations are properly initialized."""
        config = RunConfiguration()
        assert hasattr(config, "output")
        assert hasattr(config, "logging")



"""A script to infer the version number from the metadata."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("gwsim_pop")
except PackageNotFoundError:
    # Fallback for source checkouts or environments without installed metadata.
    __version__ = "0+unknown"



"""Tests for YAML utility functions."""

from __future__ import annotations

from enum import Enum
from pathlib import Path

import pytest
import yaml

from gwsim_pop.utils.yaml import read_yaml, write_yaml


class Color(Enum):
    """An Enum for testing."""

    RED = "red"
    """Red color."""


class TestYAMLUtils:
    """Test cases for YAML utility functions."""

    def test_read_yaml_valid_file(self, tmp_path: Path) -> None:
        """Test reading a valid YAML file."""
        # Create a temporary YAML file
        yaml_content = """
        key1: value1
        key2: 42
        key3:
          nested: value
        """
        test_file = tmp_path / "test.yaml"
        test_file.write_text(yaml_content)

        # Read the file
        result = read_yaml(test_file)
        expected = {
            "key1": "value1",
            "key2": 42,
            "key3": {"nested": "value"},
        }
        assert result == expected

    def test_read_yaml_invalid_file(self, tmp_path: Path) -> None:
        """Test reading an invalid YAML file."""
        # Create a temporary invalid YAML file
        test_file = tmp_path / "invalid.yaml"
        test_file.write_text("invalid: [yaml")

        # This should raise a YAML error
        with pytest.raises(yaml.YAMLError):
            read_yaml(test_file)

    def test_write_yaml_simple(self, tmp_path: Path) -> None:
        """Test writing simple YAML data without round_trip."""
        test_file = tmp_path / "output.yaml"
        data = {"key1": "value1", "key2": 42, "key3": [1, 2, 3]}

        write_yaml(test_file, data)

        # Read back the file to verify
        result = read_yaml(test_file)
        assert result == data

    def test_write_yaml_with_round_trip(self, tmp_path: Path) -> None:
        """Test writing YAML data with round_trip enabled."""
        test_file = tmp_path / "output_round_trip.yaml"
        data = {"key1": "value1", "key2": 42, "key3": [1, 2, 3]}

        write_yaml(test_file, data, round_trip=True)

        # Read back the file to verify
        result = read_yaml(test_file)
        assert result == data

    def test_write_yaml_with_enum(self, tmp_path: Path) -> None:
        """Test writing YAML data containing enums."""
        test_file = tmp_path / "enum.yaml"
        # Test enum is defined in the yaml module, but for testing, let's just use a simple dict
        data = {"enum_key": "value1", "normal_key": "value"}

        write_yaml(test_file, data)

        # Read back the file to verify
        result = read_yaml(test_file)
        assert result == {"enum_key": "value1", "normal_key": "value"}

    def test_write_yaml_with_enum_round_trip(self, tmp_path: Path) -> None:
        """Test writing YAML data containing enums with round_trip."""
        test_file = tmp_path / "enum_round_trip.yaml"
        # Test enum is defined in the yaml module, but for testing, let's just use a simple dict
        data = {"enum_key": Color.RED, "normal_key": "value"}

        write_yaml(test_file, data, round_trip=True)

        # Read back the file to verify
        result = read_yaml(test_file)
        assert result == {"enum_key": Color.RED.value, "normal_key": "value"}

    def test_write_yaml_directory_creation(self, tmp_path: Path) -> None:
        """Test that write_yaml creates parent directories if they don't exist."""
        nested_file = tmp_path / "nested" / "deep" / "test.yaml"
        data = {"key": "value"}

        write_yaml(nested_file, data)

        # Verify that the file exists and can be read
        assert nested_file.exists()
        result = read_yaml(nested_file)
        assert result == data

    def test_write_yaml_preserves_ordering_with_round_trip(self, tmp_path: Path) -> None:
        """Test that round_trip preserves key ordering."""
        test_file = tmp_path / "ordered.yaml"
        data = {
            "first": "value1",
            "second": "value2",
            "third": "value3",
        }

        write_yaml(test_file, data, round_trip=True)

        # Read back the file to verify
        result = read_yaml(test_file)
        assert result == data

    def test_read_yaml_empty_file(self, tmp_path: Path) -> None:
        """Test reading an empty YAML file."""
        test_file = tmp_path / "empty.yaml"
        test_file.write_text("")

        with pytest.raises(ValueError, match="Expected a YAML mapping at the top level of"):
            _result = read_yaml(test_file)

    def test_write_yaml_empty_data(self, tmp_path: Path) -> None:
        """Test writing empty data."""
        test_file = tmp_path / "empty_data.yaml"
        data = {}

        write_yaml(test_file, data)

        result = read_yaml(test_file)
        assert result == data

    def test_write_yaml_none_data(self, tmp_path: Path) -> None:
        """Test writing None data."""
        test_file = tmp_path / "none_data.yaml"
        data = {"key": None}

        write_yaml(test_file, data)

        result = read_yaml(test_file)
        assert result == {"key": None}



"""Tests for the advanced configuration."""

from __future__ import annotations

import pytest
from pydantic import ValidationError

from gwsim_pop.config.advanced import AdvancedConfiguration


class TestAdvancedConfiguration:
    """Test cases for AdvancedConfiguration."""

    def test_default_values(self) -> None:
        """Test that the default values are correctly set."""
        config = AdvancedConfiguration()
        assert config.backend == "jax"
        assert config.jit is True
        assert config.vectorization == "vmap"
        assert config.device == "auto"

    def test_explicit_values(self) -> None:
        """Test that explicit values are correctly set."""
        config = AdvancedConfiguration(backend="jax", jit=False, vectorization="manual loop", device="cpu")
        assert config.backend == "jax"
        assert config.jit is False
        assert config.vectorization == "manual loop"
        assert config.device == "cpu"

    def test_backend_validation_success(self) -> None:
        """Test that jax backend validation passes."""
        config = AdvancedConfiguration(backend="jax")
        assert config.backend == "jax"

    def test_backend_validation_failure(self) -> None:
        """Test that numpy backend validation fails as expected."""
        with pytest.raises(ValidationError) as exc_info:
            AdvancedConfiguration(backend="numpy")

        assert "backend='numpy' is not implemented yet" in str(exc_info.value)

    def test_field_descriptions(self) -> None:
        """Test that field descriptions are correctly set."""
        config = AdvancedConfiguration()
        # Access field descriptions through model_fields
        fields = config.__class__.model_fields

        assert fields["backend"].description == "Backend for the computation."
        assert fields["jit"].description == "Whether to jit critical functions."
        assert fields["vectorization"].description == "Vectorization method."
        assert fields["device"].description == "Device to use for the run."

    def test_all_valid_backends(self) -> None:
        """Test that all valid backend values work correctly."""
        # Test that jax is valid
        config = AdvancedConfiguration(backend="jax")
        assert config.backend == "jax"

    def test_partial_configuration(self) -> None:
        """Test that partial configuration works correctly."""
        config = AdvancedConfiguration(jit=False)
        assert config.backend == "jax"
        assert config.jit is False
        assert config.vectorization == "vmap"
        assert config.device == "auto"

    def test_configuration_from_dict(self) -> None:
        """Test that configuration can be created from a dictionary."""
        config_dict = {"backend": "jax", "jit": True, "vectorization": "pmap", "device": "gpu"}
        config = AdvancedConfiguration(**config_dict)
        assert config.backend == "jax"
        assert config.jit is True
        assert config.vectorization == "pmap"
        assert config.device == "gpu"



"""Tests for the output configuration."""

from __future__ import annotations

import pytest
from pydantic import ValidationError

from gwsim_pop.config.output import OutputConfiguration


class TestOutputConfiguration:
    """Test cases for OutputConfiguration."""

    def test_default_values(self) -> None:
        """Test that default values are correctly set."""
        config = OutputConfiguration()
        assert config.directory == "."
        assert config.format == "hdf5"
        assert config.compression == "gzip"
        assert config.overwrite is False
        assert config.save_metadata is True

    def test_explicit_values(self) -> None:
        """Test that explicit values are correctly set."""
        config = OutputConfiguration(
            directory="/custom/path", format="npz", compression="lzf", overwrite=True, save_metadata=False
        )
        assert config.directory == "/custom/path"
        assert config.format == "npz"
        assert config.compression == "lzf"
        assert config.overwrite is True
        assert config.save_metadata is False

    def test_directory_field_description(self) -> None:
        """Test that directory field has correct description."""
        config = OutputConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["directory"].description == "Output directory."

    def test_format_field_description(self) -> None:
        """Test that format field has correct description."""
        config = OutputConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["format"].description == "Storage format."

    def test_compression_field_description(self) -> None:
        """Test that compression field has correct description."""
        config = OutputConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["compression"].description == "Compression filter."

    def test_overwrite_field_description(self) -> None:
        """Test that overwrite field has correct description."""
        config = OutputConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["overwrite"].description == "Whether or not to overwrite existing files."

    def test_save_metadata_field_description(self) -> None:
        """Test that save_metadata field has correct description."""
        config = OutputConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["save_metadata"].description == "Whether or not to save metadata."

    def test_configuration_from_dict(self) -> None:
        """Test that configuration can be created from a dictionary."""
        config_dict = {
            "directory": "/custom/path",
            "format": "csv",
            "compression": "szip",
            "overwrite": True,
            "save_metadata": False,
        }
        config = OutputConfiguration(**config_dict)
        assert config.directory == "/custom/path"
        assert config.format == "csv"
        assert config.compression == "szip"
        assert config.overwrite is True
        assert config.save_metadata is False

    def test_valid_format_values(self) -> None:
        """Test that all valid format values work correctly."""
        valid_formats = ["hdf5", "npz", "csv"]

        for fmt in valid_formats:
            config = OutputConfiguration(format=fmt)
            assert config.format == fmt

    def test_valid_compression_values(self) -> None:
        """Test that all valid compression values work correctly."""
        valid_compressions = [None, "gzip", "lzf", "szip"]

        for comp in valid_compressions:
            config = OutputConfiguration(compression=comp)
            assert config.compression == comp

    def test_directory_default_value(self) -> None:
        """Test that directory default value is correct."""
        config = OutputConfiguration()
        assert config.directory == "."

    def test_format_default_value(self) -> None:
        """Test that format default value is correct."""
        config = OutputConfiguration()
        assert config.format == "hdf5"

    def test_compression_default_value(self) -> None:
        """Test that compression default value is correct."""
        config = OutputConfiguration()
        assert config.compression == "gzip"

    def test_overwrite_default_value(self) -> None:
        """Test that overwrite default value is correct."""
        config = OutputConfiguration()
        assert config.overwrite is False

    def test_save_metadata_default_value(self) -> None:
        """Test that save_metadata default value is correct."""
        config = OutputConfiguration()
        assert config.save_metadata is True

    def test_partial_configuration(self) -> None:
        """Test that partial configuration works correctly."""
        config = OutputConfiguration(directory="/custom", overwrite=True)
        assert config.directory == "/custom"
        assert config.format == "hdf5"  # default
        assert config.compression == "gzip"  # default
        assert config.overwrite is True
        assert config.save_metadata is True  # default

    def test_empty_configuration(self) -> None:
        """Test that empty configuration works correctly."""
        config = OutputConfiguration()
        assert config.directory == "."
        assert config.format == "hdf5"
        assert config.compression == "gzip"
        assert config.overwrite is False
        assert config.save_metadata is True

    def test_invalid_format_value(self) -> None:
        """Test that invalid format values raise ValidationError."""
        with pytest.raises(ValidationError):
            OutputConfiguration(format="invalid_format")

    def test_invalid_compression_value(self) -> None:
        """Test that invalid compression values raise ValidationError."""
        with pytest.raises(ValidationError):
            OutputConfiguration(compression="invalid_compression")



"""Tests for the logging configuration."""

from __future__ import annotations

import pytest
from pydantic import ValidationError

from gwsim_pop.config.logging import LoggingConfiguration
from gwsim_pop.utils.log import LoggingLevel


class TestLoggingConfiguration:
    """Test cases for LoggingConfiguration."""

    def test_default_values(self) -> None:
        """Test that default values are correctly set."""
        config = LoggingConfiguration()
        assert config.level == LoggingLevel.INFO
        assert config.filename == "simulation.log"

    def test_explicit_values(self) -> None:
        """Test that explicit values are correctly set."""
        config = LoggingConfiguration(level=LoggingLevel.DEBUG, filename="custom.log")
        assert config.level == LoggingLevel.DEBUG
        assert config.filename == "custom.log"

    def test_level_field_description(self) -> None:
        """Test that level field has correct description."""
        config = LoggingConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["level"].description == "Logging level."

    def test_filename_field_description(self) -> None:
        """Test that filename field has correct description."""
        config = LoggingConfiguration()
        # Access field description through model_fields
        fields = config.__class__.model_fields
        assert fields["filename"].description == "Name of the log file."

    def test_configuration_from_dict(self) -> None:
        """Test that configuration can be created from a dictionary."""
        config_dict = {"level": "DEBUG", "filename": "test.log"}
        config = LoggingConfiguration(**config_dict)
        assert config.level == LoggingLevel.DEBUG
        assert config.filename == "test.log"

    def test_invalid_level_value(self) -> None:
        """Test that invalid level values raise ValidationError."""
        # Test with a level that is not in LoggingLevel enum
        with pytest.raises(ValidationError):
            LoggingConfiguration(level="INVALID_LEVEL")

    def test_level_enum_values(self) -> None:
        """Test that all valid level enum values work correctly."""
        # Test that all enum values can be used
        valid_levels = [
            LoggingLevel.DEBUG,
            LoggingLevel.INFO,
            LoggingLevel.WARNING,
            LoggingLevel.ERROR,
            LoggingLevel.CRITICAL,
        ]

        for level in valid_levels:
            config = LoggingConfiguration(level=level)
            assert config.level == level

    def test_filename_default_value(self) -> None:
        """Test that filename default value is correct."""
        config = LoggingConfiguration()
        assert config.filename == "simulation.log"

    def test_filename_custom_value(self) -> None:
        """Test that custom filename values work correctly."""
        config = LoggingConfiguration(filename="custom.log")
        assert config.filename == "custom.log"

    def test_partial_configuration(self) -> None:
        """Test that partial configuration works correctly."""
        config = LoggingConfiguration(level=LoggingLevel.WARNING)
        assert config.level == LoggingLevel.WARNING
        assert config.filename == "simulation.log"

    def test_empty_configuration(self) -> None:
        """Test that empty configuration works correctly."""
        config = LoggingConfiguration()
        assert config.level == LoggingLevel.INFO
        assert config.filename == "simulation.log"



"""Tests for smoothing distribution functions."""

import jax.numpy as jnp

from gwsim_pop.core.distributions.smoothing import log_planck_tapering_function, planck_tapering_function


class TestPlanckTaperingFunction:
    """Tests for planck_tapering_function."""

    def test_basic_functionality(self):
        """Test basic functionality of the tapering function."""
        # Test with simple values
        x = jnp.array([0.5, 1.0, 1.5, 2.0, 2.5])
        x_min = 1.0
        delta = 1.0

        result = planck_tapering_function(x, x_min, delta)
        assert result.shape == x.shape

        # Check that values are between 0 and 1
        assert jnp.all(result >= 0.0)
        assert jnp.all(result <= 1.0)

    def test_below_transition_region(self):
        """Test values below the transition region."""
        x = jnp.array([0.5])
        x_min = 1.0
        delta = 1.0

        result = planck_tapering_function(x, x_min, delta)
        assert result[0] == 0.0

    def test_above_transition_region(self):
        """Test values above the transition region."""
        x = jnp.array([2.5])
        x_min = 1.0
        delta = 1.0

        result = planck_tapering_function(x, x_min, delta)
        assert result[0] == 1.0

    def test_at_transition_region_edges(self):
        """Test values at the edges of the transition region."""
        x = jnp.array([1.0, 2.0])  # edges of [x_min, x_min + delta]
        x_min = 1.0
        delta = 1.0

        result = planck_tapering_function(x, x_min, delta)
        assert result[0] == 0.0  # At x_min, should be 0
        assert result[1] == 1.0  # At x_min + delta, should be 1

    def test_array_input(self):
        """Test with multiple values in array."""
        x = jnp.array([0.5, 1.0, 1.25, 1.5, 2.0, 2.5])
        x_min = 1.0
        delta = 1.0

        result = planck_tapering_function(x, x_min, delta)
        assert len(result) == len(x)

        # Verify boundaries
        assert result[0] == 0.0  # Below transition
        assert result[-1] == 1.0  # Above transition


class TestLogPlanckTaperingFunction:
    """Tests for log_planck_tapering_function."""

    def test_basic_functionality(self):
        """Test basic functionality of the log tapering function."""
        x = jnp.array([0.5, 1.0, 1.5, 2.0, 2.5])
        x_min = 1.0
        delta = 1.0

        result = log_planck_tapering_function(x, x_min, delta)
        assert result.shape == x.shape

    def test_below_transition_region(self):
        """Test values below the transition region."""
        x = jnp.array([0.5])
        x_min = 1.0
        delta = 1.0

        result = log_planck_tapering_function(x, x_min, delta)
        assert jnp.isinf(result[0])
        assert result[0] < 0

    def test_above_transition_region(self):
        """Test values above the transition region."""
        x = jnp.array([2.5])
        x_min = 1.0
        delta = 1.0

        result = log_planck_tapering_function(x, x_min, delta)
        assert result[0] == 0.0

    def test_at_transition_region_edges(self):
        """Test values at the edges of the transition region."""
        x = jnp.array([1.0, 2.0])  # edges of [x_min, x_min + delta]
        x_min = 1.0
        delta = 1.0

        result = log_planck_tapering_function(x, x_min, delta)
        # At x_min, should be -inf
        assert jnp.isinf(result[0])
        assert result[0] < 0
        assert result[1] == 0.0  # At x_min + delta, should be 0

    def test_array_input(self):
        """Test with multiple values in array."""
        x = jnp.array([0.5, 1.0, 1.25, 1.5, 2.0, 2.5])
        x_min = 1.0
        delta = 1.0

        result = log_planck_tapering_function(x, x_min, delta)
        assert len(result) == len(x)

        # Verify boundaries
        # Below transition
        assert jnp.isinf(result[0])
        assert result[0] < 0
        assert result[-1] == 0.0  # Above transition


class TestEdgeCases:
    """Tests for edge cases in both tapering functions."""

    def test_negative_delta(self):
        """Test with negative delta."""
        x = jnp.array([0.5, 1.0, 1.5])
        x_min = 1.0
        delta = -1.0

        # These should work but produce strange results
        result1 = planck_tapering_function(x, x_min, delta)
        result2 = log_planck_tapering_function(x, x_min, delta)
        assert result1.shape == x.shape
        assert result2.shape == x.shape

    def test_extreme_values(self):
        """Test with extreme input values."""
        x = jnp.array([0.0, 1e-10, 1.0, 1000.0, 1e10])
        x_min = 1.0
        delta = 1.0

        result1 = planck_tapering_function(x, x_min, delta)
        result2 = log_planck_tapering_function(x, x_min, delta)

        assert result1.shape == x.shape
        assert result2.shape == x.shape

    def test_small_values(self):
        """Test with very small input values."""
        x = jnp.array([1e-10, 1e-5, 1.0])
        x_min = 1.0
        delta = 1.0

        result1 = planck_tapering_function(x, x_min, delta)
        result2 = log_planck_tapering_function(x, x_min, delta)

        assert result1.shape == x.shape
        assert result2.shape == x.shape

    def test_large_values(self):
        """Test with large input values."""
        x = jnp.array([1000.0, 2000.0, 3000.0])
        x_min = 1.0
        delta = 1.0

        result1 = planck_tapering_function(x, x_min, delta)
        result2 = log_planck_tapering_function(x, x_min, delta)

        assert result1.shape == x.shape
        assert result2.shape == x.shape



"""Tests for population configuration."""

import pytest
from pydantic import ValidationError

from gwsim_pop.config.population import PopulationArguments, PopulationConfiguration


class TestPopulationArguments:
    """Tests for PopulationArguments class."""

    def test_initialization(self):
        """Test that PopulationArguments can be initialized."""
        args = PopulationArguments()
        assert isinstance(args, PopulationArguments)


class TestPopulationConfiguration:
    """Tests for PopulationConfiguration class."""

    def test_initialization_without_arguments(self):
        """Test that PopulationConfiguration can be initialized without arguments."""
        config = PopulationConfiguration(model="test_model")
        assert config.model == "test_model"
        assert isinstance(config.arguments, PopulationArguments)

    def test_initialization_with_arguments(self):
        """Test that PopulationConfiguration can be initialized with arguments."""
        # Test with default arguments
        args = PopulationArguments(custom_param="value", extra_value=42)
        config = PopulationConfiguration(model="test_model", arguments=args)
        assert config.model == "test_model"
        assert isinstance(config.arguments, PopulationArguments)
        assert config.arguments.custom_param == "value"

    def test_model_validation(self):
        """Test that model validation works correctly."""
        with pytest.raises(ValidationError):
            PopulationConfiguration()

    def test_valid_model_name(self):
        """Test that valid model names work."""
        config = PopulationConfiguration(model="valid_model_name")
        assert config.model == "valid_model_name"



"""Configuration for the selection."""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field, field_validator


class SelectionConfiguration(BaseModel):
    """Configuration for the selection."""

    enabled: bool = Field(default=False, description="Whether to apply detectability.")
    """Whether to apply detectability."""

    method: Literal["snr"] = Field(default="snr", description="Method to apply detectability.")
    """Method to apply detectability."""

    arguments: dict = Field(default_factory=dict, description="Arguments for the method to apply detectability.")
    """Arguments for the method to apply detectability."""

    # TODO: After implementing this feature, remove this validator and remove `exclude=True` in main.py.
    @field_validator("enabled")
    @classmethod
    def validate_enabled(cls, value: bool) -> bool:
        """Validate the field 'enabled'.

        Args:
            value: Value of 'enabled'.

        Returns:
            Value of 'enabled'.

        Raises:
            ValueError: If value is True.
        """
        if value is True:
            raise ValueError("The selection feature is not available yet. Set it to 'False' and rerun.")
        return value



"""Models for the configuration."""

from __future__ import annotations

from pathlib import Path

from pydantic import BaseModel, ConfigDict, Field

from gwsim_pop.config.advanced import AdvancedConfiguration
from gwsim_pop.config.cosmology import CosmologyConfiguration
from gwsim_pop.config.post_processing import PostProcessingConfiguration
from gwsim_pop.config.run import RunConfiguration
from gwsim_pop.config.selection import SelectionConfiguration
from gwsim_pop.utils.yaml import read_yaml, write_yaml


class MainConfiguration(BaseModel):
    """Configuration of simulation."""

    model_config = ConfigDict(extra="forbid")
    """Forbid the extra fields."""

    config_version: str = Field(default="0.1.0", description="Version of the configuration file.")
    """Version of the configuration file."""

    run: RunConfiguration = Field(default_factory=RunConfiguration, description="Configuration for the run.")
    """Configuration for the run."""

    cosmology: CosmologyConfiguration = Field(
        default_factory=CosmologyConfiguration, description="Configuration for the cosmology."
    )
    """Configuration for the cosmology."""

    selection: SelectionConfiguration = Field(
        default_factory=SelectionConfiguration, description="Configuration for the selection.", exclude=True
    )
    """Configuration for the selection."""

    post_processing: PostProcessingConfiguration = Field(
        default_factory=PostProcessingConfiguration, description="Configuration for the post-processing.", exclude=True
    )
    """Configuration for the post-processing."""

    advanced: AdvancedConfiguration = Field(
        default_factory=AdvancedConfiguration, description="Configuration for the advanced features."
    )
    """Configuration for the advanced features."""

    @classmethod
    def from_file(cls, filename: str | Path, encoding: str = "utf-8") -> MainConfiguration:
        """Read from file.

        Args:
            filename: File name.
            encoding: File encoding.

        Returns:
            Configuration.

        """
        data = read_yaml(filename=filename, encoding=encoding)
        return cls(**data)

    def to_file(
        self,
        filename: str | Path,
        encoding: str = "utf-8",
        exclude_none: bool = True,
        exclude_defaults: bool = False,
        round_trip: bool = False,
        **kwargs,
    ) -> None:
        """Write to file.

        Args:
            filename: File name.
            encoding: Encoding of the file.
            exclude_none: Exclude None entries.
            exclude_defaults: Exclude entries set to default.
            round_trip: If True, use ruyaml to preserve comments/order.
            **kwargs: Extra arguments passed to yaml.safe_dump (or ruyaml.YAML.dump).
        """
        data = self.model_dump(mode="python", exclude_none=exclude_none, exclude_defaults=exclude_defaults)
        write_yaml(filename=filename, data=data, encoding=encoding, round_trip=round_trip, **kwargs)



"""Configuration for the post-processing."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field, field_validator


class HookArguments(BaseModel):
    """Arguments for the hook."""

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True, validate_default=True)
    """Allow additional arguments."""


class HookConfiguration(BaseModel):
    """Configuration for the hook."""

    name: str = Field(..., description="Name of the hook.")
    """Name of the hook."""

    arguments: HookArguments = Field(default_factory=HookArguments, description="Arguments for the hook.")
    """Arguments for the hook."""


class PostProcessingConfiguration(BaseModel):
    """Configuration for post-processing."""

    hooks: list[HookConfiguration] = Field(
        default_factory=list, description="A list of functions to run in post-processing."
    )
    """A list of functions to run in post-processing."""

    # TODO: After implementing this feature, remove this validator and remove `exclude=True` in main.py.
    @field_validator("hooks")
    @classmethod
    def validate_hooks(cls, value: list[HookConfiguration]) -> list[HookConfiguration]:
        """Validate the hooks input.

        Args:
            value: Input list of hooks.

        Returns:
            A list of hook configuration.
        """
        if len(value) > 0:
            raise ValueError("Post-processing hook is not implemented yet.")
        return value



"""Generate API reference pages automatically with improved formatting."""

from __future__ import annotations

from pathlib import Path

import mkdocs_gen_files

PACKAGE_NAME = "python_package_template"

src = Path(__file__).parent.parent / "src"
root_dir = Path(__file__).parent.parent

# Copy CONTRIBUTING.md to docs directory
contributing_src = root_dir / "CONTRIBUTING.md"
contributing_dest = Path("CONTRIBUTING.md")
if contributing_src.exists():
    with open(contributing_src, encoding="utf-8") as src_fd:
        content = src_fd.read()
    with mkdocs_gen_files.open("CONTRIBUTING.md", "w") as fd:
        fd.write(content)
    mkdocs_gen_files.set_edit_path(Path("CONTRIBUTING.md"), contributing_src)

# Copy SECURITY.md to docs directory
security_src = root_dir / "SECURITY.md"
security_dest = Path("SECURITY.md")
if security_src.exists():
    with open(security_src, encoding="utf-8") as src_fd:
        content = src_fd.read()
    with mkdocs_gen_files.open("SECURITY.md", "w") as fd:
        fd.write(content)
    mkdocs_gen_files.set_edit_path(Path("SECURITY.md"), security_src)

# Generate the index page
with mkdocs_gen_files.open("reference/index.md", "w") as fd:
    fd.write("# API Reference\n\n")
    fd.write("Complete API documentation for the package.\n\n")
    fd.write("## Modules\n\n")

    # Collect unique subpackages
    modules = set()
    package_dir = src / PACKAGE_NAME
    if package_dir.exists():
        for item in package_dir.iterdir():
            if item.is_dir() and (item / "__init__.py").exists() and not item.name.startswith("_"):
                # Skip private modules
                modules.add(item.name)

    # Generate module list
    for module in sorted(modules):
        fd.write(f"- [`{PACKAGE_NAME}.{module}`]({PACKAGE_NAME}/{module}/index.md)\n")

# Process all Python modules
for path in sorted(src.rglob("*.py")):
    module_path = path.relative_to(src).with_suffix("")
    doc_path = path.relative_to(src).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1] == "__main__":
        continue

    # Skip private modules, test files, and CLI module
    if any(part.startswith("_") and part != "__init__" for part in parts):
        continue
    if parts and parts[-1].startswith("test_"):
        continue
    if "cli" in parts:  # Skip CLI module
        continue

    # Skip empty parts (from root __init__)
    if not parts:
        continue

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        indent = ".".join(parts)

        # Generate improved page content
        fd.write(f"# `{indent}`\n\n")
        fd.write("::: " + indent + "\n")
        fd.write("    options:\n")
        fd.write("      docstring_style: google\n")
        fd.write("      show_source: true\n")
        fd.write("      show_root_heading: true\n")
        fd.write("      show_object_full_path: true\n")
        fd.write("      members_order: source\n")
        fd.write("      filters:\n")
        fd.write("        - '!^_'\n")

    mkdocs_gen_files.set_edit_path(full_doc_path, path)



"""Utility functions for handling YAML files."""

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Any

import yaml
from ruyaml import YAML


# Register the representers for PyYAML
def _yaml_enum_representer(dumper, data: Enum):
    """Define the representer for Enum.

    Args:
        dumper: YAML dumper.
        data: Enum data.

    Returns:
        Representer of Enum.
    """
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data.value))


_yaml_representers = ((Enum, _yaml_enum_representer),)


def read_yaml(filename: str | Path, encoding: str = "utf-8") -> dict[str, Any]:
    """Read from file.

    Args:
        filename: File name.
        encoding: File encoding.

    Returns:
        A dictionary of data.

    """
    with Path(filename).open("r", encoding=encoding) as f:
        data = yaml.safe_load(f)
    if not isinstance(data, dict):
        raise ValueError(f"Expected a YAML mapping at the top level of '{filename}', got {type(data).__name__!r}.")
    return data


def write_yaml(
    filename: str | Path,
    data: dict[str, Any],
    encoding: str = "utf-8",
    round_trip: bool = False,
    **kwargs,
):
    """Write to file.

    Args:
        filename: File name.
        data: A dictionary of data.
        encoding: Encoding of the file.
        round_trip: If True, use ruyaml to preserve comments/order.
        **kwargs: Extra arguments passed to yaml.safe_dump (or ruyaml.YAML.dump).
    """
    path = Path(filename)

    path.parent.mkdir(parents=True, exist_ok=True)

    if round_trip:
        yaml_handler = YAML(typ="rt")
        yaml_handler.default_flow_style = False
        yaml_handler.indent(mapping=4, sequence=4, offset=2)
        # Add the representer only to this specific handler instance
        for data_type, representer in _yaml_representers:
            yaml_handler.representer.add_multi_representer(data_type, representer)

        with path.open("w", encoding=encoding) as f:
            yaml_handler.dump(data, f, **kwargs)

    else:
        # For non-round_trip, we create a copy of the data to avoid modifying the original
        # and handle enum values that need special representation
        data_copy = _prepare_data_for_yaml_dump(data)
        with path.open("w", encoding=encoding) as f:
            yaml.safe_dump(data_copy, f, sort_keys=False, allow_unicode=True, **kwargs)


def _prepare_data_for_yaml_dump(data: Any) -> Any:
    """Prepare data for YAML dumping, handling enums that might not be natively supported."""
    if isinstance(data, dict):
        return {key: _prepare_data_for_yaml_dump(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [_prepare_data_for_yaml_dump(item) for item in data]
    elif isinstance(data, Enum):
        # Convert enums to their string representation for yaml.safe_dump
        return str(data.value)
    else:
        return data



"""Configuration for the cosmology."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field


class CosmologicalArguments(BaseModel):
    """Arguments for the cosmological model."""

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True, validate_default=True)
    """Allow additional parameters."""


class CosmologyConfiguration(BaseModel):
    """Configuration for the cosmology."""

    model: str = Field(default="Planck18", description="Cosmological model.")
    """Cosmological model."""

    arguments: CosmologicalArguments = Field(
        default_factory=CosmologicalArguments, description="Arguments for the model."
    )
    """Parameters for the model."""



# ruff: noqa PL0415
"""Main entry point for the gwsim_pop CLI application."""

from __future__ import annotations

import enum
from typing import Annotated

import typer
from gwsim_pop.utils.log import LoggingLevel


# Create the main Typer app
app = typer.Typer(
    name="gwsim_pop",
    help="Main CLI for gwsim_pop.",
    rich_markup_mode="rich",
)


def setup_logging(level: LoggingLevel = LoggingLevel.INFO) -> None:
    """Set up logging with Rich handler.

    Args:
        level: Logging level.
    """
    import logging

    from rich.console import Console
    from rich.logging import RichHandler

    logger = logging.getLogger("gwsim_pop")

    logger.setLevel(level.value)

    console = Console(stderr=True)

    # Remove any existing handlers to ensure RichHandler is used
    for h in logger.handlers[:]:  # Use slice copy to avoid modification during iteration
        logger.removeHandler(h)
    # Add the RichHandler

    handler = RichHandler(
        console=console,
        rich_tracebacks=True,
        show_time=True,
        show_level=True,  # Keep level (e.g., DEBUG, INFO) for clarity
        markup=True,  # Enable Rich markup in messages for styling
        level=level.value,  # Ensure handler respects the level
        omit_repeated_times=False,
        log_time_format="%H:%M",
    )
    handler.setLevel(level.value)
    logger.addHandler(handler)

    # Prevent propagation to root logger to avoid duplicate output
    logger.propagate = False


@app.callback()
def main(
    verbose: Annotated[
        LoggingLevel,
        typer.Option("--verbose", "-v", help="Set verbosity level."),
    ] = LoggingLevel.INFO,
) -> None:
    """Main entry point for the CLI application.

    Args:
        verbose: Verbosity level for logging.
    """
    setup_logging(verbose)


def register_commands() -> None:
    """Register CLI commands."""


register_commands()



"""Test importing the gwsim_pop package and its modules."""

from __future__ import annotations

import pkgutil

import pytest

import gwsim_pop


def get_all_submodules(package):
    """Discover all submodules in the package.

    Args:
        package: The package to inspect.

    """
    submodules = []
    for _, mod_name, _ in pkgutil.walk_packages(package.__path__, package.__name__ + "."):
        submodules.append(mod_name)
    return submodules


def test_import_main_package():
    """Test that the main gwsim_pop package can be imported."""
    assert hasattr(gwsim_pop, "__version__")
    assert gwsim_pop.__version__ is not None


@pytest.mark.parametrize("module_name", get_all_submodules(gwsim_pop))
def test_import_submodule(module_name):
    """Test that all submodules can be imported."""
    __import__(module_name)
